#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_fp16.h>
#include <cublas_v2.h>
#include <ATen/cuda/CUDAContext.h>
#include <ATen/cuda/CUDAUtils.h>

// ReLU kernel using fixed iteration loops with predicated operations
// The boundary check is implemented via a ternary operator to help the compiler generate predicated instructions,
// minimizing warp divergence for threads in the last (partial) iteration.
__global__ void relu_kernel(float* input, int size) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride = gridDim.x * blockDim.x;
    int iterations = (size + stride - 1) / stride;
    for (int i = 0; i < iterations; i++) {
        int pos = idx + i * stride;
        // Use a ternary operator to load a valid value; out‐of‐bounds threads load 0.0f
        float val = (pos < size) ? input[pos] : 0.0f;
        float res = fmaxf(0.0f, val);
        if (pos < size) {
            input[pos] = res;
        }
    }
}

// Optimized 2x2 max pooling kernel with uniform control flow
// Assumes pooling window is 2x2 with stride 2. The inner max calculation uses branchless fmaxf calls.
__global__ void max_pool2d_kernel(const float* __restrict__ input, float* __restrict__ output,
    int batch_size, int channels, int height, int width, int stride) {
    // Compute output dimensions (assumes height and width are divisible by 2)
    int out_h = height / 2;
    int out_w = width / 2;
    int total = batch_size * channels * out_h * out_w;
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride_total = gridDim.x * blockDim.x;
    for (int i = idx; i < total; i += stride_total) {
         int temp = i;
         int b = temp / (channels * out_h * out_w);
         temp = temp % (channels * out_h * out_w);
         int c = temp / (out_h * out_w);
         temp = temp % (out_h * out_w);
         int ph = temp / out_w;
         int pw = temp % out_w;
         int h_offset = ph * stride;
         int w_offset = pw * stride;
         int base = ((b * channels + c) * height);
         const float* row0 = input + (base + h_offset) * width + w_offset;
         const float* row1 = row0 + width;
         float m0 = fmaxf(row0[0], row0[1]);
         float m1 = fmaxf(row1[0], row1[1]);
         output[i] = fmaxf(m0, m1);
    }
}

// Flatten kernel to copy data with uniform loop execution
__global__ void flatten_kernel(const float* input, float* output, int size) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride = gridDim.x * blockDim.x;
    int iterations = (size + stride - 1) / stride;
    for (int i = 0; i < iterations; i++) {
         int pos = idx + i * stride;
         if (pos < size) {
             output[pos] = input[pos];
         }
    }
}

// Forward function implementing a LeNet-5 architecture using custom kernels
// The conv and fc layers use PyTorch operators to ensure numerical correctness, while custom CUDA kernels
// (relu_kernel and max_pool2d_kernel) are used to reduce warp divergence in elementwise and pooling operations.
// Note: The max pooling kernel is specialized for 2x2 pooling (stride = 2).

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor conv1_weight, torch::Tensor conv1_bias,
    torch::Tensor conv2_weight, torch::Tensor conv2_bias,
    torch::Tensor fc1_weight, torch::Tensor fc1_bias,
    torch::Tensor fc2_weight, torch::Tensor fc2_bias,
    torch::Tensor fc3_weight, torch::Tensor fc3_bias
) {
    // Move inputs to CUDA and ensure contiguous layout
    x = x.to(torch::kCUDA);
    conv1_weight = conv1_weight.to(torch::kCUDA);
    conv1_bias = conv1_bias.to(torch::kCUDA);
    conv2_weight = conv2_weight.to(torch::kCUDA);
    conv2_bias = conv2_bias.to(torch::kCUDA);
    fc1_weight = fc1_weight.to(torch::kCUDA);
    fc1_bias = fc1_bias.to(torch::kCUDA);
    fc2_weight = fc2_weight.to(torch::kCUDA);
    fc2_bias = fc2_bias.to(torch::kCUDA);
    fc3_weight = fc3_weight.to(torch::kCUDA);
    fc3_bias = fc3_bias.to(torch::kCUDA);

    const int block_size = 256;
    const int num_blocks = 32;

    // First convolutional layer followed by ReLU activation
    auto conv1 = torch::conv2d(x, conv1_weight, conv1_bias, {1, 1});
    relu_kernel<<<num_blocks, block_size>>>(conv1.data_ptr<float>(), conv1.numel());

    // Custom 2x2 max pooling: allocate output tensor and apply kernel (stride = 2)
    auto pool1 = torch::empty({conv1.size(0), conv1.size(1), conv1.size(2) / 2, conv1.size(3) / 2}, conv1.options());
    max_pool2d_kernel<<<num_blocks, block_size>>>(conv1.data_ptr<float>(), pool1.data_ptr<float>(),
         conv1.size(0), conv1.size(1), conv1.size(2), conv1.size(3), 2);

    // Second convolutional layer followed by ReLU activation
    auto conv2 = torch::conv2d(pool1, conv2_weight, conv2_bias, {1, 1});
    relu_kernel<<<num_blocks, block_size>>>(conv2.data_ptr<float>(), conv2.numel());
    auto pool2 = torch::empty({conv2.size(0), conv2.size(1), conv2.size(2) / 2, conv2.size(3) / 2}, conv2.options());
    max_pool2d_kernel<<<num_blocks, block_size>>>(conv2.data_ptr<float>(), pool2.data_ptr<float>(),
         conv2.size(0), conv2.size(1), conv2.size(2), conv2.size(3), 2);

    // Flatten the pooled output
    auto flat = pool2.view({pool2.size(0), -1});

    // Fully connected layers use torch::linear to preserve high precision;
    // custom ReLU is applied between fc layers to reduce divergent branching in elementwise operations.
    auto fc1 = torch::linear(flat, fc1_weight, fc1_bias);
    relu_kernel<<<num_blocks, block_size>>>(fc1.data_ptr<float>(), fc1.numel());
    auto fc2 = torch::linear(fc1, fc2_weight, fc2_bias);
    relu_kernel<<<num_blocks, block_size>>>(fc2.data_ptr<float>(), fc2.numel());
    auto fc3 = torch::linear(fc2, fc3_weight, fc3_bias);

    return fc3;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "LeNet-5 forward pass");
}
